# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: triple_index.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='triple_index.proto',
  package='',
  syntax='proto3',
  serialized_pb=_b('\n\x12triple_index.proto\"1\n\x0fTranslationUnit\x12\x0f\n\x07\x65lement\x18\x01 \x01(\t\x12\r\n\x05index\x18\x02 \x01(\x04\"V\n\x0bTranslation\x12\"\n\x08\x65ntities\x18\x01 \x03(\x0b\x32\x10.TranslationUnit\x12#\n\trelations\x18\x02 \x03(\x0b\x32\x10.TranslationUnitb\x06proto3')
)




_TRANSLATIONUNIT = _descriptor.Descriptor(
  name='TranslationUnit',
  full_name='TranslationUnit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='element', full_name='TranslationUnit.element', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='index', full_name='TranslationUnit.index', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=71,
)


_TRANSLATION = _descriptor.Descriptor(
  name='Translation',
  full_name='Translation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='entities', full_name='Translation.entities', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='relations', full_name='Translation.relations', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=73,
  serialized_end=159,
)

_TRANSLATION.fields_by_name['entities'].message_type = _TRANSLATIONUNIT
_TRANSLATION.fields_by_name['relations'].message_type = _TRANSLATIONUNIT
DESCRIPTOR.message_types_by_name['TranslationUnit'] = _TRANSLATIONUNIT
DESCRIPTOR.message_types_by_name['Translation'] = _TRANSLATION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TranslationUnit = _reflection.GeneratedProtocolMessageType('TranslationUnit', (_message.Message,), dict(
  DESCRIPTOR = _TRANSLATIONUNIT,
  __module__ = 'triple_index_pb2'
  # @@protoc_insertion_point(class_scope:TranslationUnit)
  ))
_sym_db.RegisterMessage(TranslationUnit)

Translation = _reflection.GeneratedProtocolMessageType('Translation', (_message.Message,), dict(
  DESCRIPTOR = _TRANSLATION,
  __module__ = 'triple_index_pb2'
  # @@protoc_insertion_point(class_scope:Translation)
  ))
_sym_db.RegisterMessage(Translation)


# @@protoc_insertion_point(module_scope)
